# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Common classes and utilities."""

import logging
import inspect
import warnings
import weakref

from unidist.config.backends.mpi.envvars import MpiSpawn
from unidist.core.backends.mpi.utils import ImmutableDict

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

from unidist.core.backends.common.data_id import DataID, is_data_id
from unidist.config import MpiLog, MpiSharedObjectStore

# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py import MPI  # noqa: E402


class Operation:
    """
    Class that describes supported operations.

    Attributes
    ----------
    EXECUTE : int, default 1
        Execute a remote task.
    GET : int, default 2
        Return local data to a requester.
    PUT_DATA : int, default 3
        Save the data to a local storage.
    PUT_OWNER : int, default 4
        Save the data location to a local storage.
    PUT_SHARED_DATA : int, default 5
        Save the data into shared memory.
    WAIT : int, default 6
        Return readiness signal of a local data to a requester.
    ACTOR_CREATE : int, default 7
        Create local actor instance.
    ACTOR_EXECUTE : int, default 8
        Execute method of a local actor instance.
    CLEANUP : int, default 9
        Cleanup local object storage for out-of-scope IDs.
    TASK_DONE : int, default 10
        Increment global task counter.
    GET_TASK_COUNT : int, default 11
        Return global task counter to a requester.
    RESERVE_SHARED_MEMORY : int, default 12
        Reserve area in shared memory for the data.
    REQUEST_SHARED_DATA : int, default 13
        Return the area in shared memory with the requested data.
    CANCEL : int, default 14
        Send a message to a worker to exit the event loop.
    READY_TO_SHUTDOWN : int, default 15
        Send a message to monitor from a worker,
        which is ready to shutdown.
    SHUTDOWN : int, default 16
        Send a message from monitor to a worker to shutdown.
    """

    ### --- Master/worker operations --- ###
    EXECUTE = 1
    GET = 2
    PUT_DATA = 3
    PUT_OWNER = 4
    PUT_SHARED_DATA = 5
    WAIT = 6
    ACTOR_CREATE = 7
    ACTOR_EXECUTE = 8
    CLEANUP = 9
    ### --- Monitor operations --- ###
    TASK_DONE = 10
    GET_TASK_COUNT = 11
    RESERVE_SHARED_MEMORY = 12
    REQUEST_SHARED_DATA = 13
    ### --- Common operations --- ###
    CANCEL = 14
    READY_TO_SHUTDOWN = 15
    SHUTDOWN = 16


class MPITag:
    """
    Class that describes tags that are used internally for communications.

    Attributes
    ----------
    OPERATION : int, default: 111
        The tag for send/recv of an operation type.
    OBJECT : int, default: 112
        The tag for send/recv of a regular Python object.
    BUFFER : int, default: 113
        The tag for send/recv of a buffer-like object.
    OBJECT_BLOCKING : int, default: 114
        The tag for send/recv of a regular Python object
        to indicate the blocking get request.
    """

    OPERATION = 111
    OBJECT = 112
    BUFFER = 113
    OBJECT_BLOCKING = 114


class MetadataPackage(ImmutableDict):
    """
    The class defines metadata packages for a communication.

    Attributes
    ----------
    LOCAL_DATA : int, default: 0
        Package type indicating that the data will be sent from the local object store.
    SHARED_DATA : int, default: 1
        Package type indicating that the data will be sent from the shared object store.
    TASK_DATA : int, default: 2
        Package type indicating a task or an actor (actor method) to be sent.
    """

    LOCAL_DATA = 0
    SHARED_DATA = 1
    TASK_DATA = 2

    @classmethod
    def get_local_info(cls, data_id, s_data_len, raw_buffers_len, buffer_count):
        """
        Get information package for sending local data.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        s_data_len : int
            Main buffer length.
        raw_buffers_len : list
            A list of ``PickleBuffer`` lengths.
        buffer_count : list
            List of the number of buffers for each object
            to be serialized/deserialized using the pickle 5 protocol.

        Returns
        -------
        dict
            The information package.
        """
        return MetadataPackage(
            {
                "package_type": MetadataPackage.LOCAL_DATA,
                "id": data_id,
                "s_data_len": s_data_len,
                "raw_buffers_len": raw_buffers_len,
                "buffer_count": buffer_count,
            }
        )

    @classmethod
    def get_shared_info(
        cls, data_id, s_data_len, raw_buffers_len, buffer_count, service_index
    ):
        """
        Get information package for sending data using shared memory.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        s_data_len : int
            Main buffer length.
        raw_buffers_len : list
            A list of ``PickleBuffer`` lengths.
        buffer_count : list
            List of the number of buffers for each object
            to be serialized/deserialized using the pickle 5 protocol.
        service_index : int
            Service index in shared memory.

        Returns
        -------
        dict
            The information package.
        """
        return MetadataPackage(
            {
                "package_type": MetadataPackage.SHARED_DATA,
                "id": weakref.proxy(data_id),
                "s_data_len": s_data_len,
                "raw_buffers_len": tuple(raw_buffers_len),
                "buffer_count": tuple(buffer_count),
                "service_index": service_index,
            }
        )

    @classmethod
    def get_task_info(cls, s_data_len, raw_buffers_len, buffer_count):
        """
        Get information package for sending a task or an actor (actor method).

        Parameters
        ----------
        s_data_len : int
            Main buffer length.
        raw_buffers_len : list
            A list of ``PickleBuffer`` lengths.
        buffer_count : list
            List of the number of buffers for each object
            to be serialized/deserialized using the pickle 5 protocol.

        Returns
        -------
        dict
            The information package.
        """
        return MetadataPackage(
            {
                "package_type": MetadataPackage.TASK_DATA,
                "s_data_len": s_data_len,
                "raw_buffers_len": raw_buffers_len,
                "buffer_count": buffer_count,
            }
        )


default_class_properties = dir(type("dummy", (object,), {}))
# Mapping between operations and their names (e.g., Operation.EXECUTE: "EXECUTE")
operations_dict = dict(
    (value, name)
    for name, value in inspect.getmembers(Operation)
    if name not in default_class_properties
)


def get_op_name(op):
    """
    Get string operation name.

    Parameters
    ----------
    op : unidist.core.backends.mpi.core.common.Operation
        Operation type.

    Returns
    -------
    str
        String operation name.

    Raises
    ------
    KeyError
        If the operation does not match either of `operations_dict`.
    """
    op_name = operations_dict.get(op, None)
    if op_name is None:
        raise KeyError(f"Got unsupported operation `{op}`")
    return op_name


class MpiDataID(DataID):
    """
    Class for tracking data IDs of MPI processes.

    The class extends ``unidist.core.backends.common.data_id.DataID`` functionality,
    ensuring the uniqueness of the object to correctly construct/reconstruct it.
    Otherwise, we would get into wrong garbage collection.

    Parameters
    ----------
    owner_rank : int
        The rank of the process that owns the data.
    data_number : int
        Unique data number for the owner process.
    gc : unidist.core.backends.mpi.core.executor.GarbageCollector or None
        Local garbage collector reference.
        The actual object is for the data id owner, otherwise, ``None``.
    """

    _instances = weakref.WeakValueDictionary()

    def __new__(cls, owner_rank, data_number, gc=None):
        key = (owner_rank, data_number)
        if key in cls._instances:
            return cls._instances[key]
        else:
            new_instance = super().__new__(cls)
            cls._instances[key] = new_instance
            return new_instance

    def __init__(self, owner_rank, data_number, gc=None):
        super().__init__(f"rank_{owner_rank}_id_{data_number}")
        self.owner_rank = owner_rank
        self.data_number = data_number
        self._gc = gc

    def __getnewargs__(self):
        """
        Prepare arguments to reconstruct the object upon unpickling.

        Returns
        -------
        tuple
            Tuple of the owner rank and data number to be passed into `__new__`.
        """
        return (self.owner_rank, self.data_number)

    def __getstate__(self):
        """
        Remove a reference to garbage collector for correct `pickle` serialization.

        Returns
        -------
        dict
            State of the object without garbage collector.
        """
        state = self.__dict__.copy()
        # we remove this attribute for correct serialization,
        # as well as to reduce the length of the serialized data
        if hasattr(self, "_gc"):
            del state["_gc"]
        return state

    def __del__(self):
        """Track object deletion by garbage collector."""
        # check for existence of `._gc` attribute as
        # it is missing upon unpickling
        if hasattr(self, "_gc") and self._gc is not None:
            self._gc.collect((self.owner_rank, self.data_number))


def get_logger(logger_name, file_name, activate=None):
    """
    Configure logger and get it's instance.

    Parameters
    ----------
    logger_name : str
        Name of a logger.
    file_name : str
        File name.
    activate : optional, default: None
        Whether to enable logging or not.
        If ``None``, the value is superseded with ``MpiLog``.

    Returns
    -------
    object
        A Python logger object.
    """
    logger = logging.getLogger(logger_name)
    if not logger.hasHandlers():
        f_format = logging.Formatter("%(message)s")
        f_handler = logging.FileHandler(file_name, delay=True)
        f_handler.setFormatter(f_format)
        logger.addHandler(f_handler)

    if activate is None:
        activate = MpiLog.get()
    if activate:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.NOTSET)

    return logger


def unwrapped_data_ids_list(args):
    """
    Transform all found data ID objects to their underlying value in an iterable object.

    Parameters
    ----------
    args : iterable
        Sequence of data.

    Returns
    -------
    list
        Transformed list with underlying ``DataID`` values.
    """
    if args is None:
        return [None]
    elif is_data_id(args):
        return [args._id]
    else:
        return [arg._id if is_data_id(arg) else None for arg in args]


def materialize_data_ids(data_ids, unwrap_data_id_impl, is_pending=False):
    """
    Traverse iterable object and materialize all data IDs.

    Find all ``unidist.core.backends.mpi.core.common.MpiDataID`` instances and call `unwrap_data_id_impl` on them.

    Parameters
    ----------
    data_ids : iterable
        Iterable objects to transform recursively.
    unwrap_data_id_impl : callable
        Function to get the ID associated data from the local object store if available.
    is_pending : bool, default: False
        Status of data materialization attempt as a flag.

    Returns
    -------
    iterable or bool
        Transformed iterable object (task arguments) and status if all ``DataID`` instances were transformed.
    """

    def _unwrap_data_id(*args):
        nonlocal is_pending
        value, progress = unwrap_data_id_impl(*args)
        if not is_pending:
            is_pending = progress
        return value

    if type(data_ids) in (list, tuple, dict):
        container = type(data_ids)()
        for value in data_ids:
            unwrapped_value, is_pending = materialize_data_ids(
                data_ids[value] if isinstance(data_ids, dict) else value,
                unwrap_data_id_impl,
                is_pending=is_pending,
            )
            if isinstance(container, list):
                container += [_unwrap_data_id(unwrapped_value)]
            elif isinstance(container, tuple):
                container += (_unwrap_data_id(unwrapped_value),)
            elif isinstance(container, dict):
                container.update({value: _unwrap_data_id(unwrapped_value)})
        return container, is_pending
    else:
        unwrapped = _unwrap_data_id(data_ids)
        return unwrapped, is_pending


def check_mpich_version(target_version):
    """
    Check if the using MPICH version is equal to or greater than the target version.

    Parameters
    ----------
    target_version : str
        Required version of the MPICH library.

    Returns
    -------
    bool
        True ot false.
    """

    def versiontuple(v):
        return tuple(map(int, (v.split("."))))

    mpich_version = (
        [
            raw
            for raw in MPI.Get_library_version().split("\n")
            if "MPICH Version:" in raw
        ][0]
        .replace("\t", " ")
        .split(" ")[-1]
    )
    return versiontuple(mpich_version) >= versiontuple(target_version)


def is_shared_memory_supported(raise_warning=False):
    """
    Check if the unidist on MPI supports shared memory.

    Parameters
    ----------
    raise_warning: bool, default: False
        Whether to raise a warning or not.
        ``True`` is passed only for root process
        to have the only warning.

    Returns
    -------
    bool
        True or False.

    Notes
    -----
    Prior to the MPI 3.0 standard there is no support for shared memory.
    """
    if not MpiSharedObjectStore.get():
        return False

    if MPI.VERSION < 3:
        if raise_warning:
            warnings.warn(
                f"Shared object store for MPI backend is not supported for MPI version {MPI.VERSION} "
                "since it doesn't support shared memory feature."
            )
        return False

    # Mpich shared memory does not work with spawned processes prior to version 4.2.0.
    if (
        "MPICH" in MPI.Get_library_version()
        and MpiSpawn.get()
        and not check_mpich_version("4.2.0")
    ):
        if raise_warning:
            warnings.warn(
                "Shared object store for MPI backend is not supported in C/W model for MPICH version less than 4.2.0. "
                + "Read more about this issue in the `troubleshooting` page of the unidist documentation."
            )
        return False

    return True



import pathlib
from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import sys
import versioneer


# https://github.com/modin-project/unidist/issues/324
ray_deps = ["ray[default]>=1.13.0", "pydantic<2"]
dask_deps = ["dask[complete]>=2.22.0", "distributed>=2.22.0"]
mpi_deps = ["mpi4py>=3.0.3", "msgpack>=1.0.0", "psutil"]
if sys.version_info[1] < 8:
    mpi_deps += "pickle5"
all_deps = ray_deps + dask_deps + mpi_deps

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

_memory = Extension(
    "unidist.core.backends.mpi.core._memory",
    ["unidist/core/backends/mpi/core/memory/_memory.pyx"],
    language="c++",
    extra_compile_args=["-std=c++11"],
    extra_link_args=["-std=c++11"],
)

setup(
    name="unidist",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Unified Distributed Execution",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    packages=find_packages(),
    url="https://github.com/modin-project/unidist",
    install_requires=["packaging", "cloudpickle"],
    extras_require={
        # can be installed by pip install unidist[ray]
        "ray": ray_deps,
        "dask": dask_deps,
        "mpi": mpi_deps,
        "all": all_deps,
    },
    python_requires=">=3.9",
    ext_modules=cythonize([_memory]),
)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities which can be used for unidist behavior tuning."""

from .backends.common import Backend, CpuCount
from .backends.ray import (
    RayGpuCount,
    IsRayCluster,
    RayRedisAddress,
    RayRedisPassword,
    RayObjectStoreMemory,
)
from .backends.dask import DaskMemoryLimit, IsDaskCluster, DaskSchedulerAddress
from .backends.mpi import (
    MpiSpawn,
    MpiHosts,
    MpiPickleThreshold,
    MpiBackoff,
    MpiLog,
    MpiSharedObjectStore,
    MpiSharedObjectStoreMemory,
    MpiSharedServiceMemory,
    MpiSharedObjectStoreThreshold,
    MpiRuntimeEnv,
)
from .parameter import ValueSource

__all__ = [
    "Backend",
    "CpuCount",
    "RayGpuCount",
    "IsRayCluster",
    "RayRedisAddress",
    "RayRedisPassword",
    "RayObjectStoreMemory",
    "DaskMemoryLimit",
    "IsDaskCluster",
    "DaskSchedulerAddress",
    "MpiSpawn",
    "MpiHosts",
    "ValueSource",
    "MpiPickleThreshold",
    "MpiBackoff",
    "MpiLog",
    "MpiSharedObjectStore",
    "MpiSharedObjectStoreMemory",
    "MpiSharedServiceMemory",
    "MpiSharedObjectStoreThreshold",
    "MpiRuntimeEnv",
]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""An implementation of ``Backend`` interface using Ray."""

import ray
from ray.util import get_node_ip_address

from unidist.core.backends.ray.actor import RayActor
from unidist.core.backends.ray.remote_function import RayRemoteFunction
from unidist.core.base.backend import Backend


class RayBackend(Backend):
    """The class that implements the interface in ``Backend`` using Ray."""

    @staticmethod
    def make_remote_function(function, num_cpus, num_returns, resources):
        """
        Define a remote function.

        function : callable
            Function to be a remote function.
        num_cpus : int
            The number of CPUs to reserve for the remote function.
        num_returns : int
            The number of ``ObjectRef``-s returned by the remote function invocation.
        resources : dict
            Custom resources to reserve for the remote function.

        Returns
        -------
        RayRemoteFunction
        """
        return RayRemoteFunction(function, num_cpus, num_returns, resources)

    @staticmethod
    def make_actor(cls, num_cpus, resources):
        """
        Define an actor class.

        cls : object
            Class to be an actor class.
        num_cpus : int
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict
            Custom resources to reserve for the lifetime of the actor.

        Returns
        -------
        RayActor
            The actor class type to create.
        list
            The list of arguments for ``RayActor`` constructor.
        """
        return RayActor, [cls, num_cpus, resources]

    @staticmethod
    def get(object_refs):
        """
        Get a remote object or a list of remote objects
        from distributed memory.

        Parameters
        ----------
        object_refs : ray.ObjectRef or list of ray.ObjectRef-s
            Ray ObjectRef or a list of Ray ObjectRef objects to get data from.

        Returns
        -------
        object
            A Python object or a list of Python objects.
        """
        return ray.get(object_refs)

    @staticmethod
    def put(data):
        """
        Put `data` into distributed memory.

        Parameters
        ----------
        data : object
            Data to be put.

        Returns
        -------
        ray.ObjectRef
            Ray ObjectRef matching to data.
        """
        return ray.put(data)

    @staticmethod
    def wait(object_refs, num_returns=1):
        """
        Wait until `object_refs` are finished.

        This method returns two lists. The first list consists of
        object refs that correspond to objects that completed computations.
        The second list corresponds to the rest of the object refs (which may or may not be ready).

        Parameters
        ----------
        object_refs : list of ray.ObjectRef-s
            List of Ray ObjectRefs to be waited.
        num_returns : int, default: 1
            The number of Ray ObjectRefs that should be returned as ready.

        Returns
        -------
        tuple
            List of object refs that are ready and list of the remaining object refs.
        """
        return ray.wait(object_refs, num_returns=num_returns)

    @staticmethod
    def get_ip():
        """
        Get node IP address.

        Returns
        -------
        str
            Node IP address.
        """
        return get_node_ip_address()

    @staticmethod
    def num_cpus():
        """
        Get the number of CPUs used by the execution backend.

        Returns
        -------
        int
        """
        return int(ray.cluster_resources()["CPU"])

    @staticmethod
    def cluster_resources():
        """
        Get resources of Ray cluster.

        Returns
        -------
        dict
            Dictionary with cluster nodes info in the form
            `{"node_ip0": {"CPU": x0}, "node_ip1": {"CPU": x1}, ...}`.
        """
        node_resources = {
            node_info["NodeManagerAddress"]: {
                "CPU": int(node_info["Resources"].get("CPU", 1))
            }
            for node_info in ray.nodes()
        }

        return node_resources

    @staticmethod
    def shutdown():
        """Shutdown Ray execution backend."""
        ray.shutdown()

    def is_initialized(self):
        """
        Check if Ray backend has already been initialized.

        Returns
        -------
        bool
            True or False.
        """
        return ray.is_initialized()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Monitoring process."""

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.monitor.shared_memory_manager import (
    SharedMemoryManager,
)
from unidist.core.backends.mpi.core.shared_object_store import SharedObjectStore

# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py import MPI  # noqa: E402


mpi_state = communication.MPIState.get_instance()
logger_name = "monitor_{}".format(mpi_state.global_rank if mpi_state is not None else 0)
log_file = "{}.log".format(logger_name)
monitor_logger = common.get_logger(logger_name, log_file)


class TaskCounter:
    __instance = None

    def __init__(self):
        self.task_counter = 0

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``TaskCounter``.

        Returns
        -------
        TaskCounter
        """
        if cls.__instance is None:
            cls.__instance = TaskCounter()
        return cls.__instance

    def increment(self):
        """Increment task counter by one."""
        self.task_counter += 1


class DataIDTracker:
    """
    Class that keeps track of all completed (ready) data IDs.
    """

    __instance = None

    def __init__(self):
        self.completed_data_ids = set()

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``DataIDTracker``.

        Returns
        -------
        DataIDTracker
        """
        if cls.__instance is None:
            cls.__instance = DataIDTracker()
        return cls.__instance

    def add_to_completed(self, data_ids):
        """
        Add the given data IDs to the set of completed (ready) data IDs.

        Parameters
        ----------
        data_ids : list
            List of data IDs to be added to the set of completed (ready) data IDs.
        """
        self.completed_data_ids.update(data_ids)


class WaitHandler:
    """
    Class that handles wait requests.
    """

    __instance = None

    def __init__(self):
        self.completed_data_ids = set()
        self.awaited_data_ids = []
        self.ready = []
        self.num_returns = 0

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``WaitHandler``.

        Returns
        -------
        WaitHandler
        """
        if cls.__instance is None:
            cls.__instance = WaitHandler()
        return cls.__instance

    def add_wait_request(self, awaited_data_ids, num_returns):
        """
        Add a wait request for a list of data IDs and the number of data IDs to be awaited.

        Parameters
        ----------
        awaited_data_ids : list
            List of data IDs to be awaited.
        num_returns : int
            The number of ``DataID``-s that should be returned as ready.
        """
        self.awaited_data_ids = awaited_data_ids
        self.num_returns = num_returns

    def process_wait_requests(self):
        """
        Check if wait requests are pending to be processed.

        Process pending wait requests and send the data_ids to the requester
        if number of data IDs that are ready are equal to the num_returns.
        """
        data_id_tracker = DataIDTracker.get_instance()
        i = 0
        if self.awaited_data_ids:
            while i < len(self.awaited_data_ids):
                data_id = self.awaited_data_ids[i]
                if data_id in data_id_tracker.completed_data_ids:
                    self.ready.append(data_id)
                    self.awaited_data_ids.remove(data_id)
                    if len(self.ready) == self.num_returns:
                        operation_data = {
                            "ready": self.ready,
                            "not_ready": self.awaited_data_ids,
                        }
                        communication.mpi_send_object(
                            communication.MPIState.get_instance().global_comm,
                            operation_data,
                            communication.MPIRank.ROOT,
                        )
                        self.ready = []
                        self.awaited_data_ids = []
                else:
                    i += 1


def monitor_loop():
    """
    Infinite monitor operations processing loop.

    Tracks the number of executed tasks and completed (ready) data IDs.

    Notes
    -----
    The loop exits on special cancelation operation.
    ``unidist.core.backends.mpi.core.common.Operations`` defines a set of supported operations.
    """
    task_counter = TaskCounter.get_instance()
    mpi_state = communication.MPIState.get_instance()
    wait_handler = WaitHandler.get_instance()
    data_id_tracker = DataIDTracker.get_instance()
    shared_store = SharedObjectStore.get_instance()
    shm_manager = SharedMemoryManager()

    # Barrier to check if monitor process is ready to start the communication loop
    mpi_state.global_comm.Barrier()
    monitor_logger.debug("Monitor loop started")
    # Once all workers excluding ``Root`` and ``Monitor`` ranks are ready to shutdown,
    # ``Monitor` sends the shutdown signal to every worker, as well as notifies ``Root`` that
    # it can exit the program.
    workers_ready_to_shutdown = []
    shutdown_workers = False
    while True:
        # Listen receive operation from any source
        operation_type, source_rank = communication.mpi_recv_operation(
            mpi_state.global_comm
        )
        monitor_logger.debug(
            f"common.Operation processing - {operation_type} from {source_rank} rank"
        )
        # Proceed the request
        if operation_type == common.Operation.TASK_DONE:
            task_counter.increment()
            output_data_ids = communication.mpi_recv_object(
                mpi_state.global_comm, source_rank
            )
            data_id_tracker.add_to_completed(output_data_ids)
            wait_handler.process_wait_requests()
        elif operation_type == common.Operation.WAIT:
            # TODO: WAIT request can be received from several workers,
            # but not only from master. Handle this case when requested.
            operation_data = communication.mpi_recv_object(
                mpi_state.global_comm, source_rank
            )
            awaited_data_ids = operation_data["data_ids"]
            num_returns = operation_data["num_returns"]
            wait_handler.add_wait_request(awaited_data_ids, num_returns)
            wait_handler.process_wait_requests()
        elif operation_type == common.Operation.GET_TASK_COUNT:
            # We use a blocking send here because the receiver is waiting for the result.
            communication.mpi_send_object(
                mpi_state.global_comm,
                task_counter.task_counter,
                source_rank,
            )
        elif operation_type == common.Operation.RESERVE_SHARED_MEMORY:
            request = communication.mpi_recv_object(mpi_state.global_comm, source_rank)
            reservation_info = shm_manager.get(request["id"])
            if reservation_info is None:
                reservation_info = shm_manager.put(request["id"], request["size"])
                is_first_request = True
            else:
                is_first_request = False

            communication.mpi_send_object(
                mpi_state.global_comm,
                data={**reservation_info, "is_first_request": is_first_request},
                dest_rank=source_rank,
            )
        elif operation_type == common.Operation.REQUEST_SHARED_DATA:
            info_package = communication.mpi_recv_object(
                mpi_state.global_comm, source_rank
            )
            data_id = info_package["id"]
            if data_id is None:
                raise ValueError("Requested DataID is None")
            reservation_info = shm_manager.get(data_id)
            if reservation_info is None:
                raise RuntimeError(f"The monitor does not know the data id {data_id}")
            sh_buf = shared_store.get_shared_buffer(
                reservation_info["first_index"], reservation_info["last_index"]
            )
            communication.mpi_send_buffer(
                mpi_state.global_comm,
                sh_buf,
                dest_rank=source_rank,
                data_type=MPI.BYTE,
            )
        elif operation_type == common.Operation.CLEANUP:
            cleanup_list = communication.recv_serialized_data(
                mpi_state.global_comm, source_rank
            )
            cleanup_list = [common.MpiDataID(*tpl) for tpl in cleanup_list]
            shm_manager.clear(cleanup_list)
        elif operation_type == common.Operation.READY_TO_SHUTDOWN:
            workers_ready_to_shutdown.append(source_rank)
            shutdown_workers = len(workers_ready_to_shutdown) == len(mpi_state.workers)
        elif operation_type == common.Operation.SHUTDOWN:
            SharedObjectStore.get_instance().finalize()
            if not MPI.Is_finalized():
                MPI.Finalize()
            break  # leave event loop and shutdown monitoring
        else:
            raise ValueError(f"Unsupported operation: {operation_type}")

        if shutdown_workers:
            for rank_id in mpi_state.workers + mpi_state.monitor_processes:
                if rank_id != mpi_state.global_rank:
                    communication.mpi_send_operation(
                        mpi_state.global_comm,
                        common.Operation.SHUTDOWN,
                        rank_id,
                    )
            communication.mpi_send_object(
                mpi_state.global_comm,
                common.Operation.SHUTDOWN,
                communication.MPIRank.ROOT,
            )
            SharedObjectStore.get_instance().finalize()
            if not MPI.Is_finalized():
                MPI.Finalize()
            break  # leave event loop and shutdown monitoring



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities specific for Dask backend which can be used for unidist behavior tuning."""

from .envvars import DaskMemoryLimit, IsDaskCluster, DaskSchedulerAddress

__all__ = ["DaskMemoryLimit", "IsDaskCluster", "DaskSchedulerAddress"]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

import sys
import pytest
import gc

import unidist
from unidist.config import Backend, CpuCount
from unidist.core.base.common import BackendName
from .utils import assert_equal, TestActor

unidist.init()


@pytest.fixture(autouse=True)
def call_gc_collect():
    """
    Collect all references from the previous test in order for MPI backend to work correctly.
    """
    yield
    # This is only needed for the MPI backend
    if Backend.get() == BackendName.MPI:
        gc.collect()


@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
@pytest.mark.parametrize("is_default_constructor", [True, False])
def test_actor_constructor(is_default_constructor):
    actor = TestActor.remote() if is_default_constructor else TestActor.remote(5)
    assert_equal(actor.get_accumulator.remote(), 0 if is_default_constructor else 5)


@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
def test_chaining():
    object_ref = unidist.put(7)
    actor = TestActor.remote()
    assert_equal(actor.task.remote(object_ref), 8)


@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
def test_num_returns():
    actor = TestActor.remote(7)
    object_ref0, object_ref1 = actor.multiple_returns.options(num_returns=2).remote(3)
    assert_equal(object_ref0, 3)
    assert_equal(object_ref1, 7)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="`pymp` backend incorrectly frees grabbed actors. Details are in https://github.com/modin-project/unidist/issues/65.",
)
@pytest.mark.skipif(
    sys.platform == "win32" and Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
@pytest.mark.parametrize("is_use_options", [True, False])
def test_address_space(is_use_options):
    actor0 = TestActor.options().remote(0) if is_use_options else TestActor.remote(0)
    actor1 = TestActor.options().remote(1) if is_use_options else TestActor.remote(1)

    object_ref0 = actor0.get_accumulator.remote()
    object_ref1 = actor1.get_accumulator.remote()

    assert_equal(object_ref0, 0)
    assert_equal(object_ref1, 1)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Proper serialization/deserialization is not implemented yet for pymp",
)
def test_global_capture():
    actor = TestActor.remote(0)

    @unidist.remote
    def foo():
        object_ref = actor.get_accumulator.remote()
        return unidist.get(object_ref)

    object_ref = foo.remote()

    assert_equal(object_ref, 0)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Proper serialization/deserialization is not implemented yet for pymp",
)
def test_direct_capture():
    actor = TestActor.remote(0)

    @unidist.remote
    def foo(actor):
        object_ref = actor.get_accumulator.remote()
        return unidist.get(object_ref)

    object_ref = foo.remote(actor)

    assert_equal(object_ref, 0)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Details are in https://github.com/modin-project/unidist/issues/70.",
)
def test_return_none():
    actor = TestActor.remote()
    assert_equal(actor.task_return_none.remote(), None)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Run of a remote task inside of an actor method is not implemented yet for pymp",
)
@pytest.mark.skipif(
    Backend.get() == BackendName.DASK,
    reason="Dask hungs when the number of tasks exceeds the number of cores",
)
def test_actor_scheduling():
    actor = TestActor.remote()

    @unidist.remote
    def f():
        return unidist.get(actor.get_accumulator.remote())

    # Use a larger number of cores deliberately to check the assignment
    for _ in range(CpuCount.get()):
        assert_equal(f.remote(), 0)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Actor specific functionality using Python Sequential backend."""

import unidist.core.backends.pyseq.core as py
from unidist.core.base.actor import Actor, ActorMethod
from unidist.core.base.object_ref import ObjectRef


class PySeqActorMethod(ActorMethod):
    """
    The class implements the interface in ``ActorMethod`` using Python Sequential backend.

    Parameters
    ----------
    cls : object
        An actor class from which method `method_name` will be called.
    method_name : str
        The name of the method to be called.
    """

    def __init__(self, cls, method_name):
        self._cls = cls
        self._method_name = method_name
        self._num_returns = 1

    def _remote(self, *args, num_returns=None, **kwargs):
        """
        Execute `self._method_name`.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the method.
        num_returns : int, optional
            Number of results to be returned. If it isn't
            provided, `self._num_returns` will be used.
        **kwargs : dict
            Keyword arguments to be passed in the method.

        Returns
        -------
        ObjectRef, list or None
            Type of returns depends on `num_returns` value:

            * if `num_returns == 1`, ``ObjectRef`` will be returned.
            * if `num_returns > 1`, list of ``ObjectRef``-s will be returned.
            * if `num_returns == 0`, ``None`` will be returned.
        """
        if num_returns is None:
            num_returns = self._num_returns

        class_method = getattr(self._cls, self._method_name)
        data_ids = py.submit(class_method, *args, num_returns=num_returns, **kwargs)

        if num_returns == 1:
            return ObjectRef(data_ids)
        elif num_returns > 1:
            return [ObjectRef(data_id) for data_id in data_ids]
        elif num_returns == 0:
            return None


class PySeqActor(Actor):
    """
    The class implements the interface in ``Actor`` using Python Sequential backend.

    Parameters
    ----------
    cls : object
        Class to be an actor class.
    num_cpus : int
        The number of CPUs to reserve for the lifetime of the actor.
    resources : dict
        Custom resources to reserve for the lifetime of the actor.
    """

    def __init__(self, cls, num_cpus, resources):
        self._cls = cls
        self._num_cpus = num_cpus
        self._resources = resources
        self._actor_handle = None

    def __getattr__(self, name):
        """
        Get the attribute `name` of the `self._cls` class.

        This method creates the ``PySeqActorMethod`` object that is responsible
        for calling method `name` of the `self._cls` class.

        Parameters
        ----------
        name : str
            Name of the method to be called.

        Returns
        -------
        PySeqActorMethod
        """
        return PySeqActorMethod(self._actor_handle, name)

    def _remote(self, *args, num_cpus=None, resources=None, **kwargs):
        """
        Create actor class, specific for Python Sequential backend, from `self._cls`.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in `self._cls` class constructor.
        num_cpus : int, optional
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict, optional
            Custom resources to reserve for the lifetime of the actor.
        **kwargs : dict
            Keyword arguments to be passed in `self._cls` class constructor.

        Returns
        -------
        PySeqActor
        """
        if num_cpus is not None or self._num_cpus is not None:
            raise NotImplementedError(
                "'num_cpus' is not supported by Python Sequential backend."
            )
        if resources is not None or self._resources is not None:
            raise NotImplementedError(
                "'resources' is not supported by Python Sequential backend."
            )

        self._actor_handle = self._cls(*args, **kwargs)
        return self
